# coding=utf-8
"""
Implementation for:
- TaggedLabelFrame: a custom LabelFrame widget that allows child widgets to be identified by tags
"""
import tkinter as tk
from tkinter import ttk

from UEVaultManager.tkgui.modules.ExtendedWidgetClasses import WidgetType, ExtendedWidget, ExtendedEntry, ExtendedText, ExtendedLabel, \
    ExtendedCheckButton, ExtendedButton
from UEVaultManager.tkgui.modules.functions import log_error, log_warning, log_debug


class TaggedLabelFrame(ttk.LabelFrame):
    """
    A custom LabelFrame widget that allows child widgets to be identified by tags.
    :param args: Args to pass to the widget
    :param kwargs: Kwargs to pass to the widget
    :return: TaggedLabelFrame instance
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tagged_child = {}
        self.pack_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': False}
        self.pack_fw_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': True}  # full width
        self.lblf_fw_options = {'ipadx': 2, 'ipady': 2, 'fill': tk.X, 'expand': True}

    def add_child(
        self,
        tag: str,
        widget_type: WidgetType.ENTRY,
        default_content=None,
        width=None,
        height=None,
        label=None,
        layout_option='',
        images_folder=None,
        add_label_before=True,
        focus_out_callback=None,
        focus_in_callback=None,
        click_on_callback=None
    ) -> None:
        """
        Adds a child widget to the LabelFrame and associates it with the given tag.
        :param tag: Tag to search for (case-insensitive)
        :param widget_type: Type of widget to add. A string value of 'text', 'checkbutton', or 'entry'
        :param width: Width of the child widget. Only used for text widgets
        :param height: Height of the child widget. Only used for text widgets
        :param label: Text to display in the child widget.
        :param default_content: Default content of the child widget
        :param layout_option: Layout options to use. Default, full width
        :param images_folder: folder for image used by some widgets
        :param add_label_before: If True, adds a label before the child widget
        :param focus_out_callback: Callback to call when the child widget loses focus
        :param focus_in_callback: Callback to call when the child widget get focus
        :param click_on_callback: Callback to call when the child widget is clicked or checked

        Note: we can not use command parameter to manage callback here because it should be transmited
        to the parent widget and in that case tag won't be available as an indentificator
        """
        tag = tag.lower()
        frame = ttk.Frame(self)
        frame.pack(**self.lblf_fw_options)
        if add_label_before:
            lbl_name = ttk.Label(frame, text=ExtendedWidget.tag_to_label(tag))
            lbl_name.pack(side=tk.LEFT, **self.pack_options)
        if widget_type == WidgetType.ENTRY:
            child = ExtendedEntry(master=frame, tag=tag, default_content=default_content, height=height, width=width)
        elif widget_type == WidgetType.TEXT:
            child = ExtendedText(master=frame, tag=tag, default_content=default_content, wrap=tk.WORD, height=height, width=width)
        elif widget_type == WidgetType.LABEL:
            child = ExtendedLabel(master=frame, tag=tag, default_content=default_content)
        elif widget_type == WidgetType.CHECKBUTTON:
            child = ExtendedCheckButton(master=frame, tag=tag, default_content=default_content, label=label, images_folder=images_folder)
        elif widget_type == WidgetType.BUTTON:
            child = ExtendedButton(master=frame, tag=tag, default_content=default_content)
        else:
            error = f'Invalid widget type: {widget_type}'
            log_error(error)
            return

        self._tagged_child[tag] = child

        layout_option = self.pack_fw_options if layout_option == '' else layout_option
        child.pack(side=tk.LEFT, **layout_option)

        if focus_out_callback is not None:
            child.bind('<FocusOut>', lambda event: focus_out_callback(event=event, tag=tag))
        if focus_in_callback is not None:
            child.bind('<FocusIn>', lambda event: focus_in_callback(event=event, tag=tag))
        if click_on_callback is not None:
            child.bind('<Button-1>', lambda event: click_on_callback(event=event, tag=tag))

    def get_child_by_tag(self, tag: str):
        """
        Returns the child widget associated with the given tag.
        :param tag: Tag to search for (case-insensitive)
        :return: Child widget
        """
        tag = tag.lower()
        return self._tagged_child.get(tag)

    def get_children(self) -> dict:
        """
        Returns the dictionary of tagged children.
        :return: A dictionary of tagged children
        """
        return self._tagged_child

    def set_default_content(self, tag='') -> None:
        """
        Sets the default content of the child widget associated with the given tag.
        :param tag: Tag to search for (case-insensitive)
        """
        tag = tag.lower()
        widget = self.get_child_by_tag(tag)
        if widget is not None:
            widget.set_content(widget.default_content)

    def set_child_values(self, tag='', content='', label=None, row=-1, col=-1) -> None:
        """
        Sets the content of the child widget associated with the given tag.
        Also sets its row and column index.
        :param tag: Tag to search for (case-insensitive)
        :param label: Text to set (only for CheckButton widget)
        :param content: Content to set
        :param row: Row index to set
        :param col: Column index to set
        """
        tag = tag.lower()
        widget = self.get_child_by_tag(tag)
        if widget is not None:
            widget.set_content(content)
            widget.row = row
            widget.col = col
            if label is not None:
                try:
                    widget.set_label(label)
                except AttributeError:
                    log_debug(f'Widget with tag {tag} does not have a set_label method')
        else:
            log_warning(f'Widget with tag {tag} not found')



# coding=utf-8
"""
implementation for:
- VersionNum: The version of the database or/and class.
- DatabaseConnection: Context manager for opening and closing a database connection.
- UEAssetDbHandler: Handles database operations for the UE Assets.
"""
import datetime
import inspect
import logging
import os
import random
import sqlite3
from enum import Enum

from faker import Faker

from UEVaultManager.core import default_datetime_format
from UEVaultManager.models.csv_data import get_sql_field_name_list
from UEVaultManager.models.UEAssetClass import UEAsset
from UEVaultManager.tkgui.modules.functions_no_deps import path_from_relative_to_absolute, convert_to_str_datetime, create_uid
from UEVaultManager.utils.cli import check_and_create_path

test_only_mode = False  # add some limitations to speed up the dev process - Set to True for debug Only


class VersionNum(Enum):
    """
    The version of the database or/and class.
    Used when checking if database must be upgraded by comparing with the class version
    """
    # when a new version is added to the VersionNum enum
    # - add code for the new version to the create_tables() method
    # - add code for the new version check to the check_and_upgrade_database() method
    V0 = 0  # invalid version
    V1 = 1  # initial version : only the "standard" marketplace columns
    V2 = 2  # add the columns used fo user data to the "standard" marketplace columns
    V3 = 3  # add the last_run table to get data about the last run of the app
    V4 = 4  # add custom_attributes field to the assets table
    V5 = 5  # future version


class DatabaseConnection:
    """
    Context manager for opening and closing a database connection
    :param database_name: The name of the database file.
    """

    def __init__(self, database_name: str):
        self.database_name = database_name
        self.conn = sqlite3.connect(self.database_name)

    def __enter__(self):
        """
        Open the database connection.
        :return: The sqlite3.Connection object.
        """
        self.conn = sqlite3.connect(self.database_name)
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Close the database connection.
        :param exc_type: Exception type.
        :param exc_val: Exception value.
        :param exc_tb: Exception traceback.
        """
        self.conn.close()
        self.conn = None


class UEAssetDbHandler:
    """
    Handles database operations for the UE Assets.
    :param database_name: The name of the database file.
    :param reset_database: If True, the database will be reset.

    Note: The database will be created if it doesn't exist.
    """

    def __init__(self, database_name: str, reset_database: bool = False):
        self.logger = logging.getLogger(__name__)
        # self.logger.setLevel(logging.DEBUG)
        self.database_name = database_name
        self.db_version = VersionNum.V0  # updated in check_and_upgrade_database()
        # the user fields that must be preserved when updating the database
        # these fields are also present in the asset table and in the UEAsset.init_data() method
        # THEY WILL BE PRESERVED when parsing the asset data
        self.user_fields = ['comment', 'stars', 'must_buy', 'test_result', 'installed_folder', 'alternative', 'origin']
        # the field we keep for previous data. NEED TO BE SEPARATED FROM self.user_fields
        # THEY WILL BE USED (BUT NOT FORCELY PRESERVED) when parsing the asset data
        self.existing_data_fields = ['id', 'price', 'old_price', 'asset_url', 'grab_result', 'date_added_in_db']
        self.existing_data_fields.extend(self.user_fields)
        self.connection = None
        self._init_connection()
        if reset_database:
            self.drop_tables()

        self.check_and_upgrade_database()

    def __del__(self):
        # log could alrready be destroyed before here
        # self.logger.debug('Deleting UEAssetDbHandler and closing connection')
        print('Deleting UEAssetDbHandler and closing connection')
        self._close_connection()

    def _close_connection(self) -> None:
        """
        Close the database connection.
        """
        if self.connection is not None:
            self.connection.close()
        self.connection = None

    def _init_connection(self) -> sqlite3.Connection:
        """
        Initialize the database connection.
        :return: The sqlite3.Connection object.

        Note: It will also set self.Connection property.
        """
        self._close_connection()  # close the connection IF IT WAS ALREADY OPENED
        try:
            self.connection = DatabaseConnection(self.database_name).conn
        except sqlite3.Error as error:
            print(f'Error while connecting to sqlite: {error!r}')

        return self.connection

    def _get_db_version(self) -> VersionNum:
        """
        Check the database version.
        :return: The database version.
        """
        try:
            if self.connection is not None:
                cursor = self.connection.cursor()
                cursor.execute("PRAGMA user_version")
                ret = int(cursor.fetchone()[0])  # type: int
                cursor.close()
                db_version = VersionNum(ret)
                return db_version
        except sqlite3.OperationalError:
            self.logger.critical('Could not get the database version')
            return VersionNum.V0

    def _set_db_version(self, new_version: VersionNum) -> None:
        """
        Set the database version.
        :param new_version: The new database version.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f'PRAGMA user_version = {new_version.value}')
            cursor.close()
            self.logger.info(f'database version is now set to {new_version}')

    def _check_db_version(self, minimal_db_version: VersionNum, caller_name='this method') -> bool:
        """
        Check if the database version is compatible with the current method.
        :param minimal_db_version: The minimal database version that is compatible with the current method.
        :param caller_name: The name of the method that called this method.
        :return:  True if the database version is compatible, otherwise False.
        """
        if self.db_version.value < minimal_db_version.value:
            message = f'This version of {caller_name} is only compatible with {minimal_db_version} or more database.\nPlease upgrade the database first.'
            self.logger.critical(message)
            return False
        return True

    def _add_missing_columns(self, table_name: str, required_columns: dict) -> None:
        """
        Add missing columns to a table.
        :param table_name: Name of the table.
        :param required_columns: Dictionary of columns to add. Key is the column name, value is the data type.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"PRAGMA table_info({table_name})")
            columns = {row[1]: row for row in cursor.fetchall()}
            for column_name, data_type in required_columns.items():
                if column_name not in columns:
                    cursor.execute(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {data_type}")
            self.connection.commit()
            cursor.close()

    def db_exists(self) -> bool:
        """
        Check if the database file exists.
        :return: True if the database file exists, otherwise False.
        """
        return os.path.exists(self.database_name)

    def create_tables(self, upgrade_to_version=VersionNum.V1) -> None:
        """
        Create the tables if they don't exist.
        :param upgrade_to_version: The database version we want to upgrade TO
        """
        # all the following steps must be run sequentially
        if upgrade_to_version.value >= VersionNum.V1.value:
            if not self.is_table_exist('assets'):
                if self.connection is not None:
                    cursor = self.connection.cursor()
                    # create the first version of the database
                    # Note:
                    # - this table has the same structure as the json files saved inside the method UEAssetScraper.save_to_file()
                    # - the order of columns must match the order of the fields in UEAsset.init_data() method
                    query = """
                    CREATE TABLE IF NOT EXISTS assets ( 
                        id TEXT PRIMARY KEY NOT NULL, 
                        namespace TEXT, 
                        catalog_item_id TEXT, 
                        title TEXT, 
                        category TEXT, 
                        author TEXT, 
                        thumbnail_url TEXT, 
                        asset_slug TEXT, 
                        currency_code TEXT, 
                        description TEXT, 
                        technical_details TEXT, 
                        long_description TEXT, 
                        tags TEXT, 
                        comment_rating_id TEXT, 
                        rating_id TEXT, 
                        status TEXT, 
                        price REAL, 
                        discount_price REAL, 
                        discount_percentage REAL, 
                        is_catalog_item BOOLEAN, 
                        is_new BOOLEAN, 
                        free BOOLEAN, 
                        discounted BOOLEAN, 
                        can_purchase BOOLEAN, 
                        owned INTEGER, 
                        review REAL, 
                        review_count INTEGER
                    )
                    """
                    cursor.execute(query)
                    self.connection.commit()
                    cursor.close()
        if upgrade_to_version.value >= VersionNum.V3.value:
            if not self.is_table_exist('last_run'):
                # Note: this table has the same structure as the data saved in the last_run_filename inside the method UEAssetScraper.save_all_to_files()
                cursor = self.connection.cursor()
                query = "CREATE TABLE IF NOT EXISTS last_run (date DATETIME, mode TEXT, files_count INTEGER, items_count INTEGER, scraped_ids TEXT)"
                cursor.execute(query)
                self.connection.commit()
                cursor.close()

    def check_and_upgrade_database(self, upgrade_from_version: VersionNum = None) -> None:
        """
        Change the tables structure according to different versions.
        :param upgrade_from_version: The version we want to upgrade FROM. if None, the current version will be used
        """
        if not self.is_table_exist('assets'):
            previous_version = VersionNum.V0
        else:
            # force an update of the db_version property
            previous_version = self._get_db_version()
            # if self.db_version == VersionNum.V0:
            #    raise Exception("Invalid database version or database is badly initialized")

        self.db_version = previous_version
        if upgrade_from_version is None:
            upgrade_from_version = self.db_version

        # all the following steps must be run sequentially
        if upgrade_from_version.value <= VersionNum.V1.value:
            # necessary steps to upgrade to version 1, aka create tables
            self.create_tables(upgrade_to_version=VersionNum.V1)
            # necessary steps to upgrade to version 2
            # add the columns used fo user data to the "standard" marketplace columns
            self._add_missing_columns(
                'assets',
                required_columns={
                    'asset_id': 'TEXT',
                    'asset_url': 'TEXT',
                    'comment': 'TEXT',
                    'stars': 'INTEGER',
                    'must_buy': 'BOOL',
                    'test_result': 'TEXT',
                    'installed_folder': 'TEXT',
                    'alternative': 'TEXT',
                    'origin': 'TEXT',
                    'page_title': 'TEXT',
                    'obsolete': 'BOOL',
                    'supported_versions': 'TEXT',
                    'creation_date': 'DATETIME',
                    'update_date': 'DATETIME',
                    'date_added_in_db': 'DATETIME',
                    'grab_result': 'TEXT',
                    'old_price': 'REAL'
                }
            )
            self.db_version = VersionNum.V2
            upgrade_from_version = self.db_version
            self.logger.info(f'Database upgraded to {upgrade_from_version}')
        if upgrade_from_version == VersionNum.V2:
            # necessary steps to upgrade to version 3
            # add the last_run table to get data about the last run of the ap
            self.create_tables(upgrade_to_version=VersionNum.V3)
            self.db_version = VersionNum.V3
            upgrade_from_version = self.db_version
            self.logger.info(f'Database upgraded to {upgrade_from_version}')
        if upgrade_from_version == VersionNum.V3:
            # necessary steps to upgrade to version 4
            self._add_missing_columns('assets', required_columns={'custom_attributes': 'TEXT'})
            self.db_version = VersionNum.V4
            upgrade_from_version = self.db_version
            self.logger.info(f'Database upgraded to {upgrade_from_version}')
        if upgrade_from_version == VersionNum.V4:
            # necessary steps to upgrade to version 5
            # does not exist yet
            # do stuff here
            # self.db_version = VersionNum.V5
            # upgrade_from_version = self.db_version
            # self.logger.info(f'Database upgraded to {upgrade_from_version}')
            pass
        if previous_version != self.db_version:
            self._set_db_version(self.db_version)

    def is_table_exist(self, table_name) -> bool:
        """
        Check if a table exists.
        :param table_name: The name of the table to check.
        :return:  True if the 'assets' table exists, otherwise False.
        """
        result = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'")
            result = cursor.fetchone()
            cursor.close()
        return result is not None

    def get_rows_count(self, table_name='assets') -> int:
        """
        Get the number of rows in the given table.
        :param table_name: The name of the table.
        :return:  The number of rows in the 'assets' table.
        """
        row_count = 0
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
            row_count = cursor.fetchone()[0]
            cursor.close()
        return row_count

    def save_last_run(self, data: dict):
        """
        Save the last run data into the 'last_run' table.
        :param data: A dictionary containing the data to save.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(VersionNum.V3, caller_name=inspect.currentframe().f_code.co_name):
            return
        if self.connection is not None:
            cursor = self.connection.cursor()
            query = "INSERT INTO last_run (date, mode, files_count, items_count, scraped_ids) VALUES (:date, :mode, :files_count, :items_count, :scraped_ids)"
            cursor.execute(query, data)
            cursor.close()
            self.connection.commit()

    def set_assets(self, assets) -> None:
        """
        Insert or update assets into the 'assets' table.
        :param assets: A dictionary or a list of dictionaries representing assets.

        NOTE: the (existing) user fields data should have already been added or merged the assets dictionary
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(VersionNum.V2, caller_name=inspect.currentframe().f_code.co_name):
            return
        if self.connection is not None:
            cursor = self.connection.cursor()
            if not isinstance(assets, list):
                assets = [assets]
            str_today = datetime.datetime.now().strftime(default_datetime_format)
            # Note: the order of columns and value must match the order of the fields in UEAsset.init_data() method
            for asset in assets:
                asset['update_date'] = str_today
                asset['creation_date'] = convert_to_str_datetime(value=asset['creation_date'], date_format=default_datetime_format)
                asset['date_added_in_db'] = convert_to_str_datetime(
                    value=asset['date_added_in_db'], date_format=default_datetime_format, default=str_today
                )
                # Generate the SQL query
                # this query will insert or update the asset if it already exists
                query = """
                    REPLACE INTO assets (
                        id, namespace, catalog_item_id, title, category, author, thumbnail_url,
                        asset_slug, currency_code, description,
                        technical_details, long_description, tags, comment_rating_id,
                        rating_id, status, price, discount_price, discount_percentage,
                        is_catalog_item, is_new, free, discounted, can_purchase,
                        owned, review, review_count, asset_id, asset_url,
                        comment, stars, must_buy, test_result, installed_folder, alternative, origin, page_title,
                        obsolete, supported_versions, creation_date, update_date, date_added_in_db, grab_result,
                        old_price, custom_attributes
                    ) VALUES (
                        :id, :namespace, :catalog_item_id, :title, :category, :author, :thumbnail_url,
                        :asset_slug, :currency_code, :description,
                        :technical_details, :long_description, :tags, :comment_rating_id,
                        :rating_id, :status, :price, :discount_price, :discount_percentage,
                        :is_catalog_item, :is_new, :free, :discounted, :can_purchase,
                        :owned, :review, :review_count, :asset_id, :asset_url,
                        :comment, :stars, :must_buy, :test_result, :installed_folder, :alternative, :origin, :page_title, 
                        :obsolete, :supported_versions, :creation_date, :update_date, :date_added_in_db, :grab_result,
                        :old_price, :custom_attributes
                    )
                """
                # Execute the SQL query
                try:
                    cursor.execute(query, asset)
                except (sqlite3.IntegrityError, sqlite3.InterfaceError) as error:
                    self.logger.error(f"Error while inserting/updating asset '{asset['id']}' (tags='{asset['tags']}': {error!r}")
        try:
            self.connection.commit()
        except (sqlite3.IntegrityError, sqlite3.InterfaceError) as error:
            self.logger.error(f'Error when committing the database changes: {error!r}')

    def get_assets_data(self, fields='*') -> dict:
        """
        Get data from all the assets in the 'assets' table.
        :param fields: list of fields to return.
        :return: dictionary {ids, rows}
        """
        if not isinstance(fields, str):
            fields = ', '.join(fields)
        row_data = {}
        if self.connection is not None:
            self.connection.row_factory = sqlite3.Row
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT {fields} FROM assets")
            for row in cursor.fetchall():
                uid = row['id']
                row_data[uid] = dict(row)
            cursor.close()
        return row_data

    def get_assets_data_for_csv(self) -> (list, list):
        """
        Get data from all the assets in the 'assets' table for a "CSV file" like format.
        :return: list(rows),list (column_names)
        """
        csv_column_names = []
        rows = []
        if self.connection is not None:
            cursor = self.connection.cursor()
            # generate column names for the CSV file using AS to rename the columns
            fields = get_sql_field_name_list(exclude_csv_only=True, return_as_string=True, add_alias=True)
            query = f"SELECT {fields} FROM assets"
            if test_only_mode:
                query += ' ORDER BY date_added_in_db DESC LIMIT 3000'
            cursor.execute(query)
            rows = cursor.fetchall()
            csv_column_names = [
                description[0] for description in cursor.description
            ]  # by using the 'AS' in the SQL query, the column names are the CSV column names
            cursor.close()

        return rows, csv_column_names

    def create_empty_row(self, return_as_string=True, empty_cell='None'):
        """
        Create an empty row in the 'assets' table.
        :param return_as_string: True to return the row as a string, False to
        :param empty_cell: The value to use for empty cells.
        :return: A row (dict) or a string representing the empty row.
        """
        result = '' if return_as_string else {}
        # add a new row to the 'assets' table
        if self.connection is not None:
            cursor = self.connection.cursor()
            # generate a unique ID and check if it already exists in the database
            uid = ''
            count = 1
            while count > 0:
                uid = create_uid()
                cursor.execute("SELECT COUNT(*) FROM assets WHERE id = ?", (uid,))
                count = cursor.fetchone()[0]
            cursor.close()
            ue_asset = UEAsset()
            ue_asset.data['id'] = uid
            ue_asset.data['asset_id'] = 'dummy_row_' + uid  # dummy unique Asset_id to avoid issue
            ue_asset.data['thumbnail_url'] = empty_cell  # avoid displaying image warning on mouse over
            self.save_ue_asset(ue_asset)
            result = str(ue_asset) if return_as_string else ue_asset
        return result

    def read_ue_asset(self, uid: str) -> UEAsset:
        """
        Read an UEAsset object from the 'assets' table.
        :param uid: The ID of the asset to get.
        :return: UEAsset object.
        """
        ue_asset = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("SELECT * assets WHERE id = ?", (uid,))
            row = cursor.fetchone()
            cursor.close()
            ue_asset = UEAsset()
            ue_asset.init_from_list(data=row)
        return ue_asset

    def save_ue_asset(self, ue_asset: UEAsset) -> None:
        """
        Save an UEAsset object to the 'assets' table.
        :param ue_asset: UEAsset object to save
        """
        self.set_assets([ue_asset.data])

    def delete_asset(self, uid: str) -> None:
        """
        Delete an asset from the 'assets' table by its ID.
        :param uid: The ID of the asset to delete.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("DELETE FROM assets WHERE id = ?", (uid,))
            self.connection.commit()
            cursor.close()

    def delete_all_assets(self) -> None:
        """
        Delete all assets from the 'assets' table.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("DELETE FROM assets WHERE 1")
            self.connection.commit()
            cursor.close()

    def update_asset(self, uid: str, column: str, value) -> None:
        """
        Update a specific column of an asset in the 'assets' table by its ID.
        :param uid: The ID of the asset to update.
        :param column: The name of the column
        :param value: The new value.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"UPDATE assets SET {column} = ? WHERE id = ?", (value, uid))
            self.connection.commit()
            cursor.close()

    def drop_tables(self) -> None:
        """
        Drop the 'assets' and 'last_run' tables.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("DROP TABLE IF EXISTS assets")
            cursor.execute("DROP TABLE IF EXISTS last_run")
            self.connection.commit()
            cursor.close()

    def generate_test_data(self, number_of_rows=1) -> None:
        """
        Generate and insert the specified number of fake assets into the 'assets' table.
        :param number_of_rows: The number of fake assets to generate and insert.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(VersionNum.V2, caller_name=inspect.currentframe().f_code.co_name):
            return
        scraped_ids = []
        fake = Faker()
        for index in range(number_of_rows):
            assets_id = fake.uuid4()
            print(f'creating test asset # {index} with id {assets_id}')
            ue_asset = UEAsset()
            # the order of values must match the order of the fields in csv_data.py/csv_sql_fields dict
            data_list = [
                fake.uuid4(),  # asset_id
                fake.sentence(),  # title
                fake.word(),  # category
                round(random.uniform(0, 5), 1),  # review
                random.randint(0, 1000),  # review_count
                fake.name(),  # author
                fake.text(),  # description
                fake.word(),  # status
                round(random.uniform(1, 100), 2),  # discount_price
                round(random.uniform(0, 100), 2),  # discount_percentage
                random.choice([0, 1]),  # discounted
                random.choice([0, 1]),  # is_new
                random.choice([0, 1]),  # free
                random.choice([0, 1]),  # can_purchase
                random.choice([0, 1]),  # owned
                random.choice([0, 1]),  # obsolete
                fake.word(),  # supported_versions
                fake.word(),  # grab_result
                round(random.uniform(1, 100), 2),  # price
                random.randint(0, 1000),  # old_price
                fake.text(),  # comment
                random.randint(1, 5),  # stars
                random.choice([0, 1]),  # must_buy
                fake.word(),  # test_result
                fake.file_path(),  # installed_folder
                fake.sentence(),  # alternative
                fake.word(),  # origin
                fake.sentence(),  # custom_attributes
                fake.sentence(),  # page_title
                fake.image_url(),  # thumbnail_url
                fake.url(),  # asset_url
                fake.date_time(),  # creation_date
                fake.date_time(),  # update_date
                fake.date_time(),  # date_added_in_db
                assets_id,  # id
                fake.word(),  # namespace
                fake.uuid4(),  # catalog_item_id
                fake.slug(),  # asset_slug
                'USD',  # currency_code
                fake.text(),  # technical_details
                fake.text(),  # long_description
                [random.randint(0, 1000), random.randint(0, 1000), random.randint(0, 1000)],  # tags
                fake.uuid4(),  # comment_rating_id
                fake.uuid4(),  # rating_id
                random.choice([0, 1]),  # is_catalog_item
            ]
            ue_asset.init_from_list(data=data_list)
            self.set_assets(assets=ue_asset.data)
            scraped_ids.append(assets_id)
        content = {
            'date': datetime.datetime.now().strftime(default_datetime_format),
            'mode': 'save',
            'files_count': 0,
            'items_count': number_of_rows,
            'scraped_ids': ','.join(scraped_ids)
        }
        self.save_last_run(content)


if __name__ == "__main__":
    # the following code is just for class testing purposes
    clean_data = True
    read_data_only = True  # if True, the code will not create fake assets, but only read them from the database

    db_folder = path_from_relative_to_absolute('../../../scraping/')
    db_name = os.path.join(db_folder, 'assets.db')
    check_and_create_path(db_name)
    asset_handler = UEAssetDbHandler(database_name=db_name, reset_database=(clean_data and not read_data_only))

    if read_data_only:
        # Read existing assets
        asset_list = asset_handler.get_assets_data()
        print("Assets:", asset_list)
    elif test_only_mode:
        # Create fake assets
        rows_to_create = 300
        if not clean_data:
            rows_count = asset_handler.get_rows_count()
            print(f"Rows count: {rows_count}")
            rows_to_create -= rows_count
        print(f"Creating {rows_to_create} rows")
        asset_handler.generate_test_data(rows_to_create)

    rows_count = asset_handler.get_rows_count()
    print(f"FINAL Rows count: {rows_count}")



# coding=utf-8
"""
implementation for:
- FakeProgressWindow: a fake ProgressWindow object to use when no ProgressWindow object is provided and provide the same interface
"""


class FakeProgressWindow:
    """
    A fake ProgressWindow object to use when no ProgressWindow object is provided and provide the same interface
    Usefull to avoid importing the ProgressWindow class and all its dependencies (all the tkinter stuff)
    """

    def __init__(self, title='Dummy'):
        self.title = title
        self.text = ''
        self.value = 0
        self.max_value = 0
        self.quit_on_close = False
        self.continue_execution = True

    def reset(self, new_title=None, new_value=None, new_text=None, new_max_value=None) -> None:
        """ FAKE METHOD"""
        self.title = new_title
        self.text = new_text
        self.value = new_value
        self.max_value = new_max_value

    def update_and_continue(self, value=0, increment=0, text=None) -> bool:
        """ FAKE METHOD"""
        if value:
            self.value = value
        else:
            self.value += increment
        self.text = text
        return True

    def close_window(self, destroy_window=True, _event=None) -> None:
        """ FAKE METHOD"""
        self.quit_on_close = destroy_window



# coding=utf-8
"""
Implementation for:
- UEAssetScraper: a class that handles scraping data from the Unreal Engine Marketplace.
"""
import concurrent.futures
import json
import logging
import os
import random
import time
from datetime import datetime
from itertools import chain
from threading import current_thread

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.api.egs import EPCAPI, GrabResult, is_asset_obsolete
from UEVaultManager.core import default_datetime_format
from UEVaultManager.models.UEAssetClass import UEAsset
from UEVaultManager.models.UEAssetDbHandlerClass import UEAssetDbHandler
from UEVaultManager.tkgui.modules.FakeProgressWindowClass import FakeProgressWindow
from UEVaultManager.tkgui.modules.functions_no_deps import check_and_get_folder, create_uid, convert_to_str_datetime, convert_to_datetime

test_only_mode = False  # add some limitations to speed up the dev process - Set to True for debug Only


def get_filename_from_asset_data(asset_data) -> str:
    """
    Returns the filename to use to save the asset data
    :param asset_data: the asset data
    :return:  the filename to use to save the asset data
    """
    try:
        app_id = asset_data['releaseInfo'][0]['appId']
        filename = f'{app_id}.json'
    except KeyError:
        uid = asset_data.get('id', None)
        if uid is None:
            uid = asset_data.get('catalogItemId', create_uid())
        filename = f'_no_appId_asset_{uid}.json'
    return filename


class UEAssetScraper:
    """
    A class that handles scraping data from the Unreal Engine Marketplace.
    It saves the data in json files and/or in a sqlite database
    :param start: An int representing the starting index for the data to retrieve. Defaults to 0.
    :param start: An int representing the ending index for the data to retrieve. Defaults to 0.
    :param assets_per_page: An int representing the number of items to retrieve per request. Defaults to 100.
    :param sort_by: A string representing the field to sort by. Defaults to 'effectiveDate'.
    :param sort_order: A string representing the sort order. Defaults to 'ASC'.
    :param timeout: A float representing the timeout for the requests. Defaults to 10.0.
    :param max_threads: An int representing the maximum number of threads to use. Defaults to 8. Set to 0 to disable multithreading.
    :param load_from_files: A boolean indicating whether to load the data from files instead of scraping it. Defaults to False. If set to True, store_in_files will be set to False and store_in_db will be set to True.
    :param store_in_files: A boolean indicating whether to store the data in csv files. Defaults to True. Could create lots of files (1 file per asset)
    :param store_in_db: A boolean indicating whether to store the data in a sqlite database. Defaults to True.
    :param store_ids: A boolean indicating whether to store and save the IDs of the assets. Defaults to False. Could be memory consuming.
    :param use_raw_format: A boolean indicating whether to store the data in a raw format (as returned by the API) or after data have been parsed. Defaults to True.
    :param clean_database: A boolean indicating whether to clean the database before saving the data. Defaults to False.
    :param engine_version_for_obsolete_assets: A string representing the engine version to use to check if an asset is obsolete.
    :param egs: An EPCAPI object (session handler). Defaults to None. If None, a new EPCAPI object will be created and the session used WON'T BE LOGGED.
    :param progress_window: A ProgressWindow object. Defaults to None. If None, a new ProgressWindow object will be created.
    """

    def __init__(
        self,
        start=0,
        stop=0,
        assets_per_page=100,
        sort_by='effectiveDate',  # other values: 'title','currentPrice','discountPercentage'
        sort_order='DESC',  # other values: 'ASC'
        timeout=10.0,
        max_threads=8,
        store_in_files=True,
        store_in_db=True,
        store_ids=False,
        use_raw_format=True,
        load_from_files=False,
        clean_database=False,
        engine_version_for_obsolete_assets=None,
        egs: EPCAPI = None,
        progress_window=None,  # don't use a typed annotation here to avoid import
    ) -> None:
        self.start = start
        self.stop = stop
        self.assets_per_page = assets_per_page
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.load_from_files = load_from_files
        self.store_in_files = store_in_files
        self.store_in_db = store_in_db
        self.store_ids = store_ids
        self.use_raw_format = use_raw_format
        self.clean_database = clean_database
        # test several ways to get the following value depending on the context
        # noinspection PyBroadException
        try:
            self.engine_version_for_obsolete_assets = (
                engine_version_for_obsolete_assets or gui_g.UEVM_cli_ref.core.engine_version_for_obsolete_assets or
                gui_g.s.engine_version_for_obsolete_assets
            )
        except Exception:
            self.engine_version_for_obsolete_assets = None

        self.last_run_filename = 'last_run.json'
        self.urls_list_filename = 'urls_list.txt'
        self.assets_data_folder = os.path.join(gui_g.s.scraping_folder, 'assets', 'marketplace')
        self.owned_assets_data_folder = os.path.join(gui_g.s.scraping_folder, 'assets', 'owned')
        self.db_name = os.path.join(gui_g.s.scraping_folder, 'assets.db')
        self.max_threads = max_threads
        self.threads_count = 0

        self.files_count = 0
        self.scraped_data = []  # the scraper scraped_data. Increased on each call to get_data_from_url(). Could be huge !!
        self.scraped_ids = []  # store IDs of all items
        self.owned_asset_ids = []  # store IDs of all owned items
        self.existing_data = {
        }  # dictionary {ids, rows} : the data in database before scraping. Contains only minimal information (user fields, id, price)

        self.urls = []  # list of all urls to scrap

        self.egs = EPCAPI(timeout=timeout) if egs is None else egs
        self.logger = logging.getLogger(__name__)
        self.asset_db_handler = UEAssetDbHandler(self.db_name)
        self.thread_executor = None
        self.thread_executor_must_stop = False

        if progress_window is None:
            progress_window = FakeProgressWindow()
        self.progress_window = progress_window

        # self.logger.setLevel(logging.DEBUG)
        if self.load_from_files:
            self.store_in_files = False
            self.store_in_db = True

        if (assets_per_page > 100) or (assets_per_page < 1):
            self.assets_per_page = 100
            self.logger.error(f'assets_per_page must be between 1 and 100. Set to 100')

        message = f'UEAssetScraper initialized with max_threads= {max_threads}, start= {start}, stop= {stop}, assets_per_page= {assets_per_page}, sort_by= {sort_by}, sort_order= {sort_order}'
        message += f'\nData will be load from files in {self.assets_data_folder}' if self.load_from_files else ''
        message += f'\nAsset Data will be saved in files in {self.assets_data_folder}' if self.store_in_files else ''
        message += f'\nOwned Asset Data will be saved in files in {self.owned_assets_data_folder}' if self.store_in_files else ''
        message += f'\nData will be saved in database in {self.db_name}' if self.store_in_db else ''
        message += f'\nAsset Ids will be saved in {self.last_run_filename} or in database' if self.store_ids else ''
        self._log_info(message)

    def _log_debug(self, message):
        """ a simple wrapper to use when cli is not initialized"""
        if gui_g.UEVM_cli_ref is None:
            print(message)
        else:
            if test_only_mode:
                # force printing debug messages in test_only_mode
                self.logger.info(message)
            else:
                self.logger.debug(message)

    def _log_info(self, message):
        """ a simple wrapper to use when cli is not initialized"""
        if gui_g.UEVM_cli_ref is None:
            print(message)
        else:
            self.logger.info(message)

    def _log_warning(self, message):
        """ a simple wrapper to use when cli is not initialized"""
        if gui_g.UEVM_cli_ref is None:
            print(message)
        else:
            self.logger.warning(message)

    def _log_error(self, message):
        """ a simple wrapper to use when cli is not initialized"""
        if gui_g.UEVM_cli_ref is None:
            print(message)
        else:
            self.logger.error(message)

    def _parse_data(self, json_data: dict = None, owned_assets_only=False) -> []:
        """
        Parses on or more asset data from the response of an url query.
        :param json_data: A dictionary containing the data to parse.
        :param owned_assets_only: if True, only the owned assets are scraped
        :return: A list containing the parsed data.
        """
        content = []
        no_text_data = ''

        if json_data is None:
            return content

        try:
            # get the list of assets after a "scraping" of the json data using URL
            assets_data_list = json_data['data']['elements']
        except KeyError:
            self._log_debug('data is not in a "scraped" format')
            # create a list of one asset when data come from a json file
            assets_data_list = [json_data]

        for asset_data in assets_data_list:
            uid = asset_data.get('id', None)
            if uid is None:
                # this should never occur
                self._log_warning(f'No id found for asset {asset_data}. Passing to next asset')
                continue
            categories = asset_data.get('categories', None)
            release_info = asset_data.get('releaseInfo', {})
            asset_existing_data = self.existing_data.get(uid, None)
            price = 0
            discount_price = 0
            discount_percentage = 0
            average_rating = 0
            rating_total = 0
            supported_versions = no_text_data
            origin = 'Marketplace'  # by default when scraped from marketplace
            date_now = datetime.now().strftime(default_datetime_format)
            grab_result = GrabResult.NO_ERROR.name

            # make some calculation with the "raw" data
            # ------------
            # set simple fields
            asset_data['thumbnail_url'] = asset_data['thumbnail']
            asset_data['category'] = categories[0]['name'] if categories else ''
            asset_data['author'] = asset_data['seller']['name']
            asset_data['asset_url'] = self.egs.get_asset_url(asset_data.get('urlSlug', None))
            try:
                asset_data['asset_id'] = release_info[0]['appId']  # latest release
            except (KeyError, AttributeError):
                grab_result = GrabResult.NO_APPID.name
                asset_data['asset_id'] = uid

            # set url and (re) update data if needed
            existing_url = asset_existing_data['asset_url'] if asset_existing_data else ''
            try:
                if existing_url and asset_existing_data and asset_data['asset_url'] != existing_url and asset_existing_data.get(
                    'grab_result', GrabResult.NO_ERROR.name
                ) != GrabResult.NO_ERROR.name:
                    self._log_warning(f'URL have changed in database and asset for asset #{uid}. Parsing data from {existing_url}')
                    # we use existing_url and not asset_data['asset_url'] because it could have been corrected by the user
                    # TODO: parse data from existing_url
            except (KeyError, TypeError) as error:
                self._log_debug(f'Error checking asset_url for asset #{uid}: {error!r}')

            # set prices and discount
            if asset_data.get('priceValue', 0) > 0:
                # tbh the logic here is flawed as hell lol. discount should only be set if there's a discount Epic wtf
                # price = asset_data['priceValue'] if asset_data['priceValue'] == asset_data['discountPriceValue'] else asset_data['discountPriceValue'] # here we keep the CURRENT price
                price = float(asset_data['priceValue'])  # here we keep the NORMAL price
                discount_price = float(asset_data['discountPriceValue'])
                # price are in cents
                price /= 100
                discount_price /= 100
                discount_percentage = 100 - int(asset_data['discountPercentage'])
                # discount_percentage = 0.0 if (discount_price == 0.0 or price == 0.0 or discount_price == price) else int((price-discount_price) / price * 100.0)

            if discount_price == 0:
                discount_price = price
            asset_data['price'] = price
            asset_data['discount_price'] = discount_price
            asset_data['discount_percentage'] = discount_percentage

            # set rating
            try:
                average_rating = asset_data['rating']['averageRating']
                rating_total = asset_data['rating']['total']
            except KeyError:
                self.logger.debug(f'No rating for {asset_data["title"]}')
            asset_data['review'] = average_rating
            asset_data['review_count'] = rating_total

            # custom attributes
            asset_data['custom_attributes'] = ''
            try:
                custom_attributes = asset_data['customAttributes']
                if custom_attributes != {}:
                    self._log_info(f'asset {uid}:custom_attributes= {custom_attributes}')
                    # has an external link
                    custom_attributes = 'external_link:' + custom_attributes['BuyLink']['value'] if custom_attributes.get('BuyLink', '') else ''
                    asset_data['custom_attributes'] = custom_attributes
            except (KeyError, AttributeError):
                asset_data['custom_attributes'] = ''

            # set origin
            asset_data['page_title'] = asset_data['title']
            try:
                tmp_list = [','.join(item.get('compatibleApps')) for item in release_info]
                supported_versions = ','.join(tmp_list)
            except TypeError as error:
                self._log_debug(f'Error getting compatibleApps for asset #{uid}: {error!r}')
            asset_data['supported_versions'] = supported_versions
            asset_data['origin'] = origin

            asset_data['update_date'] = date_now
            tmp_date = release_info[0].get('dateAdded', no_text_data) if release_info else no_text_data
            tmp_date = convert_to_datetime(tmp_date, formats_to_use=[gui_g.s.epic_datetime_format, gui_g.s.csv_datetime_format])
            tmp_date = convert_to_str_datetime(tmp_date, gui_g.s.csv_datetime_format)
            asset_data['creation_date'] = tmp_date
            tmp_date = asset_existing_data.get('date_added_in_db', None) if asset_existing_data else date_now
            asset_data['date_added_in_db'] = tmp_date

            asset_data['obsolete'] = is_asset_obsolete(supported_versions, self.engine_version_for_obsolete_assets)
            old_price = asset_existing_data.get('price', None) if asset_existing_data else 0
            older_price = asset_existing_data.get('old_price', None) if asset_existing_data else 0
            asset_data['old_price'] = old_price if old_price else older_price
            # the current parsing process does not produce as many error as the previous one
            # mainly, there is no error during the process
            old_grab_result = asset_existing_data.get('grab_result', GrabResult.NO_ERROR.name) if asset_existing_data else GrabResult.NO_ERROR.name
            if owned_assets_only and old_grab_result == GrabResult.NO_ERROR.name:
                # if no error occurs and if we only parse owned assets, the parsed data ARE less complete than the "normal" one
                # so, we set the grab result to PARTIAL
                grab_result = GrabResult.PARTIAL.name
            asset_data['grab_result'] = grab_result
            # not pertinent, kept for compatibility with CSV format. Could be removed later after checking usage
            # asset_data['app_name'] = asset_data['asset_id']
            # asset_data['developer'] = asset_data['author']
            # asset_data['compatible_versions'] = no_text_data
            # asset_data['ue_version'] = no_text_data
            # asset_data['uid'] = no_text_data

            # we use an UEAsset object to store the data and create a valid dict from it
            ue_asset = UEAsset()

            # we use copy data for user_fields to preserve user data
            if asset_existing_data:
                for field in self.asset_db_handler.user_fields:
                    old_value = asset_existing_data.get(field, None)
                    if old_value:
                        asset_data[field] = old_value

            ue_asset.init_from_dict(asset_data)
            content.append(ue_asset.data)
            message = f'Asset #{uid} added to content ue_asset.data: owned={ue_asset.data["owned"]} creation_date={ue_asset.data["creation_date"]}'
            self._log_debug(message)
            # print(message) # only if run from main
            if self.store_ids:
                try:
                    self.scraped_ids.append(uid)
                except (AttributeError, TypeError):
                    self._log_debug(f'Error adding {uid} to self.scraped_ids')
        # end for asset_data in json_data['data']['elements']:
        return content

    def _save_in_db(self, last_run_content: dict) -> None:
        """
        Stores the asset data in the database.
        """
        if not self.store_in_db:
            return
        # convert the list of ids to a string for the database only
        last_run_content['scraped_ids'] = ','.join(self.scraped_ids) if self.store_ids else ''

        if self.clean_database:
            # next line will delete all assets in the database
            self.asset_db_handler.delete_all_assets()
            self.asset_db_handler.set_assets(self.scraped_data)
        else:
            self.asset_db_handler.set_assets(self.scraped_data)
        self.asset_db_handler.save_last_run(last_run_content)

    def gather_all_assets_urls(self, empty_list_before=False, save_result=True, owned_assets_only=False) -> None:
        """
        Gathers all the URLs (with pagination) to be parsed and stores them in a list for further use.
        :param empty_list_before: if True, the list of URLs is emptied before adding the new ones
        :param save_result: if True, the list of URLs is saved in the database
        :param owned_assets_only: if True, only the owned assets are scraped
        """
        if empty_list_before:
            self.urls = []
        start_time = time.time()
        if self.stop <= 0:
            self.stop = self.egs.get_scraped_asset_count(owned_assets_only=owned_assets_only)
        assets_count = self.stop - self.start
        pages_count = int(assets_count / self.assets_per_page)
        if (assets_count % self.assets_per_page) > 0:
            pages_count += 1
        self.progress_window.reset(new_value=0, new_text='Gathering URLs', new_max_value=pages_count)
        for i in range(int(pages_count)):
            if not self.progress_window.update_and_continue(value=i, text=f'Gathering URL ({i + 1}/{pages_count})'):
                return
            start = self.start + (i * self.assets_per_page)
            if owned_assets_only:
                url = self.egs.get_owned_scrap_url(start, self.assets_per_page)
            else:
                url = self.egs.get_scrap_url(start, self.assets_per_page, self.sort_by, self.sort_order)
            self.urls.append(url)
        self._log_info(f'It took {(time.time() - start_time):.3f} seconds to gather {len(self.urls)} urls')
        if save_result:
            self.save_to_file(filename=self.urls_list_filename, data=self.urls, is_json=False, is_owned=owned_assets_only)

    def get_data_from_url(self, url='', owned_assets_only=False) -> None:
        """
        Grabs the data from the given url and stores it in the scraped_data property.
        :param url: The url to grab the data from. If not given, uses the url property of the class.
        :param owned_assets_only: if True, only the owned assets are scraped
        """
        if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.continue_execution:
            return

        if not url:
            self._log_error('No url given to get_data_from_url()')
            return

        thread_data = ''
        if self.threads_count > 1:
            # add a delay when multiple threads are used
            time.sleep(random.uniform(1.0, 3.0))
            thread = current_thread()
            thread_data = f' ==> By Thread name={thread.name}'

        self._log_info(f'--- START scraping data from {url}{thread_data}')

        json_data = self.egs.get_scraped_assets(url)
        if json_data.get('errorCode', '') != '':
            self._log_error(f'Error getting data from url {url}: {json_data["errorCode"]}')
            return
        try:
            number = len(json_data['data']['elements'])
        except KeyError:
            number = 0
        self._log_info(f'==> parsed url {url}: got {number} assets')

        if json_data:
            if self.store_in_files and self.use_raw_format:
                for asset_data in json_data['data']['elements']:
                    filename = get_filename_from_asset_data(asset_data)
                    self.save_to_file(filename=filename, data=asset_data, is_owned=owned_assets_only)
                    self.files_count += 1
            content = self._parse_data(json_data)
            self.scraped_data.append(content)

    def get_scraped_data(self, owned_assets_only=False) -> None:
        """
        Loads from files or downloads the items from the URLs and stores them in the scraped_data property.
        The execution is done in parallel using threads.
        :param owned_assets_only: if True, only the owned assets are scraped

        Note: if self.urls is None or empty, gather_urls() will be called first.
        """

        def stop_executor(tasks) -> None:
            """
            Cancel all outstanding tasks and shut down the executor
            :param tasks: tasks to cancel
            """
            for _, task in tasks.items():
                task.cancel()
            self.thread_executor.shutdown(wait=False)

        # store previous data in the existing_data property
        self.existing_data = self.asset_db_handler.get_assets_data(fields=self.asset_db_handler.existing_data_fields)

        if self.load_from_files:
            self.load_from_json_files()
        else:
            # start_time = time.time()
            # self.get_owned_asset_ids()
            # message = f'It took {(time.time() - start_time):.3f} seconds to get {len(self.owned_asset_ids)} owned asset ids'
            # self._log_info(message)

            start_time = time.time()
            if self.urls is None or len(self.urls) == 0:
                self.gather_all_assets_urls(owned_assets_only=owned_assets_only)
            self.progress_window.reset(new_value=0, new_text='Scraping data from URLs', new_max_value=len(self.urls))
            if self.max_threads > 0:
                self.threads_count = min(self.max_threads, len(self.urls))
                # threading processing COULD be stopped by the progress window
                self.progress_window.show_stop_button()
                self.thread_executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.threads_count, thread_name_prefix="Asset_Scaper")
                futures = {}
                # for url in self.urls:
                while len(self.urls) > 0:
                    url = self.urls.pop()
                    # Submit the task and add its Future to the dictionary
                    future = self.thread_executor.submit(lambda url_param: self.get_data_from_url(url_param, owned_assets_only), url)
                    futures[url] = future

                with concurrent.futures.ThreadPoolExecutor():
                    for future in concurrent.futures.as_completed(futures.values()):
                        try:
                            _ = future.result()
                            # print("Result: ", result)
                        except Exception as error:
                            self._log_warning(f'The following error occurs in threading: {error!r}')
                        if not self.progress_window.update_and_continue(increment=1):
                            # self._log_info(f'User stop has been pressed. Stopping running threads....')   # will flood console
                            stop_executor(futures)
                self.thread_executor.shutdown(wait=False)

                # # Wait for all the tasks to finish
                # concurrent.futures.wait(futures.values())
                # for key, future in futures.items():
                #     try:
                #         future.result()
                #     except Exception as error:
                #         self._log_warning(f'thread execution with key {key} has generated an exception: {error!r}')
                # self.thread_executor.shutdown(wait=False)
            else:
                for url in self.urls:
                    self.get_data_from_url(url, owned_assets_only)
            if self.store_in_files and self.use_raw_format:
                message = f'It took {(time.time() - start_time):.3f} seconds to download {len(self.urls)} urls and store the data in {self.files_count} files'
            else:
                message = f'It took {(time.time() - start_time):.3f} seconds to download {len(self.urls)} urls'
            self._log_info(message)
            # format the list to be 1 long list rather than multiple lists nested in a list - [['1'], ['2'], ...] -> ['1','2', ...]
            self.scraped_data = list(chain.from_iterable(self.scraped_data))

    def save_to_file(self, prefix='assets', filename=None, data=None, is_json=True, is_owned=False) -> bool:
        """
        Saves JSON data to a file.
        :param data: A dictionary containing the data to save. Defaults to None. If None, the data will be used.
        :param prefix: A string representing the prefix to use for the file name. Defaults to 'assets'.
        :param filename: A string representing the file name to use. Defaults to None. If None, a file name will be generated using the prefix and the start and count properties.
        :param is_json: A boolean indicating whether the data is JSON or not. Defaults to True.
        :param is_owned: A boolean indicating whether the data is owned assets or not. Defaults to False.
        :return: A boolean indicating whether the file was saved successfully.
        """
        if data is None:
            data = self.scraped_data

        if data is None or len(data) == 0:
            self._log_warning('No data to save')
            return False

        folder = self.owned_assets_data_folder if is_owned else self.assets_data_folder

        _, folder = check_and_get_folder(folder)
        if filename is None:
            filename = prefix
            if self.start > 0:
                filename += '_' + str(self.start)
                filename += '_' + str(self.start + self.assets_per_page)
            filename += '.json'
        filename = os.path.join(folder, filename)
        try:
            with open(filename, 'w') as fh:
                if is_json:
                    json.dump(data, fh)
                else:
                    # write data as a list in the file
                    fh.write('\n'.join(data))
            self._log_debug(f'Data saved into {filename}')
            return True
        except PermissionError as error:
            self._log_warning(f'The following error occured when saving data into {filename}:{error!r}')
            return False

    def load_from_json_files(self, owned_assets_only=False) -> None:
        """
        Load all JSON data retrieved from the Unreal Engine Marketplace API to paginated files.
        :param owned_assets_only: if True, only the owned assets are scraped
        """
        start_time = time.time()
        self.files_count = 0
        self.scraped_ids = []
        self.scraped_data = []
        folder = self.owned_assets_data_folder if owned_assets_only else self.assets_data_folder

        files = os.listdir(folder)
        files_count = len(files)
        # Note: this data have the same structure as the table last_run inside the method UEAsset.create_tables()
        self._log_info(f'Loading {files_count} files from {folder}')
        self.progress_window.reset(new_value=0, new_text='Loading asset data from json files', new_max_value=files_count)
        for filename in files:
            if filename.endswith('.json') and filename != self.last_run_filename:
                # self._log_debug(f'Loading {filename}')
                with open(os.path.join(folder, filename), 'r') as fh:
                    try:
                        json_data = json.load(fh)
                    except json.decoder.JSONDecodeError as error:
                        self._log_warning(f'The following error occured when loading data from {filename}:{error!r}')
                        continue
                    assets_data = self._parse_data(json_data)  # self._parse_data returns a list of assets
                    for asset_data in assets_data:
                        self.scraped_data.append(asset_data)
                self.files_count += 1
                if not self.progress_window.update_and_continue(increment=1):
                    return
                if test_only_mode and self.files_count >= 100:
                    break
        message = f'It took {(time.time() - start_time):.3f} seconds to load the data from {self.files_count} files'
        self._log_info(message)

        # save results in the last_run file
        content = {
            'date': str(datetime.now()),
            'mode': 'load_owned' if owned_assets_only else 'load',
            'files_count': self.files_count,
            'items_count': len(self.scraped_data),
            'scraped_ids': self.scraped_ids if self.store_ids else ''
        }
        filename = os.path.join(folder, self.last_run_filename)
        with open(filename, 'w') as fh:
            json.dump(content, fh)

        self.progress_window.reset(new_value=0, new_text='Saving into database', new_max_value=0)
        self._save_in_db(last_run_content=content)

    def save(self, owned_assets_only=False) -> None:
        """
        Saves all JSON data retrieved from the Unreal Engine Marketplace API to paginated files.
        :param owned_assets_only: if True, only the owned assets are scraped
        """
        if owned_assets_only:
            self._log_info('Only Owned Assets will be scraped')

        self.get_scraped_data(owned_assets_only=owned_assets_only)
        if not self.progress_window.continue_execution:
            return

        # Note: this data have the same structure as the table last_run inside the method UEAsset.create_tables()
        content = {
            'date': str(datetime.now()),
            'mode': 'save_owned' if owned_assets_only else 'save',
            'files_count': 0,
            'items_count': len(self.scraped_data),
            'scraped_ids': ''
        }

        start_time = time.time()
        if self.store_in_files and not self.use_raw_format:
            self.progress_window.reset(new_value=0, new_text='Saving into files', new_max_value=len(self.scraped_data))
            # store the data in a AFTER parsing it
            self.files_count = 0
            for asset_data in self.scraped_data:
                filename = get_filename_from_asset_data(asset_data)
                self.save_to_file(filename=filename, data=asset_data, is_owned=owned_assets_only)
                self.files_count += 1
                if not self.progress_window.update_and_continue(increment=1):
                    return
            message = f'It took {(time.time() - start_time):.3f} seconds to save the data in {self.files_count} files'
            self._log_info(message)

        # save results in the last_run file
        content['files_count'] = self.files_count
        content['scraped_ids'] = self.scraped_ids if self.store_ids else ''
        folder = self.owned_assets_data_folder if owned_assets_only else self.assets_data_folder
        filename = os.path.join(folder, self.last_run_filename)
        with open(filename, 'w') as fh:
            json.dump(content, fh)

        self.progress_window.reset(new_value=0, new_text='Saving into database', new_max_value=None)
        self._save_in_db(last_run_content=content)
        self._log_info('data saved')


if __name__ == '__main__':
    # the following code is just for class testing purposes

    # set the number of rows to retrieve per page
    # As the asset are saved individually by default, this value is only use for pagination in the files that store the url
    # it speeds up the process of requesting the asset list
    ue_asset_per_page = 100

    if test_only_mode:
        # shorter and faster list for testing only
        # disabling threading is used for debugging (fewer exceptions are raised if threads are used)
        threads = 0  # set to 0 to disable threading
        start_row = 15000
        stop_row = 15000 + ue_asset_per_page
        clean_db = False
        load_data_from_files = False
    else:
        threads = 16
        start_row = 0
        stop_row = 0  # 0 means no limit
        clean_db = True
        load_data_from_files = False  # by default the scraper will rebuild the database from scratch

    scraper = UEAssetScraper(
        start=start_row,
        stop=stop_row,
        assets_per_page=ue_asset_per_page,
        max_threads=threads,
        store_in_db=True,
        store_in_files=not load_data_from_files,
        store_ids=False,
        load_from_files=load_data_from_files,
        clean_database=clean_db
    )
    scraper.save()



# coding=utf-8
"""
implementation for:
- UEAsset:  A class to represent an Unreal Engine asset
"""
import logging

from UEVaultManager.models.csv_data import get_sql_field_name_list
from UEVaultManager.utils.cli import init_dict_from_data


class UEAsset:
    """
    A class to represent an Unreal Engine asset. With the EGS data and user data.
    :param engine_version_for_obsolete_assets: The engine version to use to check if an asset is obsolete
    """

    def __init__(self, engine_version_for_obsolete_assets=None):
        if engine_version_for_obsolete_assets is None:
            self.engine_version_for_obsolete_assets = '4.26'  # no access to the engine_version_for_obsolete_assets global settings here without importing its module
        else:
            self.engine_version_for_obsolete_assets = engine_version_for_obsolete_assets
        self.data = {}
        self.log = logging.getLogger('UEAsset')
        self.log.setLevel(logging.INFO)
        self.init_data()

    def __str__(self) -> str:
        """
        Return a string representation of the asset.
        :return: A string representation of the asset.
        """
        return ','.join(str(value) for value in self.data.values())

    def init_data(self) -> None:
        """
        Initialize the EGS data dictionary.

        Note: the keys of self.Data dict are initialized here
        """
        self.data = {key: None for key in get_sql_field_name_list(include_asset_only=True)}
        # self.data = {
        #     'id': None,
        #     'namespace': None,
        #     'catalog_item_id': None,
        #     'title': None,
        #     "category": None,
        #     'author': None,
        #     'thumbnail_url': None,
        #     'asset_slug': None,
        #     'currency_code': None,
        #     'description': None,
        #     'technical_details': None,
        #     'long_description': None,
        #     'tags': None,
        #     'comment_rating_id': None,
        #     'rating_id': None,
        #     'status': None,
        #     'price': None,
        #     'discount_price': None,
        #     'discount_percentage': None,
        #     'is_catalog_item': None,
        #     'is_new': None,
        #     'free': None,
        #     'discounted': None,
        #     'can_purchase': None,
        #     'owned': None,
        #     'review': None,
        #     'review_count': None,
        #     'asset_id': None,
        #     'asset_url': None,
        #     'comment': None,
        #     'stars': None,
        #     'must_buy': None,
        #     'test_result': None,
        #     'installed_folder': None,
        #     'alternative': None,
        #     'origin': None,
        #     'page_title': None,
        #     'obsolete': None,
        #     'supported_versions': None,
        #     'creation_date': None,
        #     'update_date': None,
        #     'date_added_in_db': None,
        #     'grab_result': None,
        #     'old_price': None,
        #     'custom_attributes': None,
        # }

    def init_from_dict(self, data: dict = None) -> None:
        """
        Initialize the asset data from the given dictionaries.
        :param data: source dictionary for the EGS data
        """
        if data:
            self.init_data()
            # copy all the keys from the data dict to the self.data dict
            init_dict_from_data(self.data, data)
            self.convert_tag_list_to_string()

    def init_from_list(self, data: list = None) -> None:
        """
        Initialize the asset data from the given lists.
        :param data: source list for the EGS data
        """
        if data:
            self.init_data()
            # fill dictionary with the data from the list
            # Note: keep in mind that the order for the values of the list and must correspond to the order of the keys in self.data
            keys = self.data.keys()
            self.data = dict(zip(keys, data))
            self.convert_tag_list_to_string()

    def convert_tag_list_to_string(self) -> None:
        """
        INPLACE Convert the tags id list of an asset_data dict to a string.
        """
        tags = self.data.get('tags', None)
        if not tags or tags == [] or tags == {}:
            tags = ''
        else:
            # TODO : find to get the value associated to each tag id. Not sure, it's possible using the API
            tags = [str(i) for i in tags]  # convert each item to a string, if not an error will be raised when joining
            tags = ','.join(tags)
        self.data['tags'] = tags
        # print(f"{self.data['asset_id']} tags converted: {tags}") # debug only
